@@ -7,5 +7,7 @@ from .snowflake import Snowflake
 from .bigquery import BigQuery
 from .redshift import Redshift
 from .presto import Presto
+from .databricks import Databricks
+from .trino import Trino
 
 from .connect import connect_to_uri
