@@ -2,7 +2,7 @@ import math
 import sys
 import logging
 from typing import Dict, Tuple, Optional, Sequence, Type, List
-from functools import lru_cache, wraps
+from functools import wraps
 from concurrent.futures import ThreadPoolExecutor
 import threading
 from abc import abstractmethod
@@ -16,6 +16,7 @@ from .database_types import (
     Float,
     ColType_UUID,
     Native_UUID,
+    String_Alphanum,
     String_UUID,
     TemporalType,
     UnknownColType,
@@ -73,7 +74,7 @@ class Database(AbstractDatabase):
 
     Used for providing connection code and implementation specific SQL utilities.
 
-    Instanciated using :meth:`~data_diff.connect_to_uri`
+    Instanciated using :meth:`~data_diff.connect`
     """
 
     TYPE_CLASSES: Dict[str, type] = {}
@@ -91,7 +92,7 @@ class Database(AbstractDatabase):
         logger.debug("Running SQL (%s): %s", type(self).__name__, sql_code)
         if getattr(self, "_interactive", False) and isinstance(sql_ast, Select):
             explained_sql = compiler.compile(Explain(sql_ast))
-            logger.info(f"EXPLAIN for SQL SELECT")
+            logger.info("EXPLAIN for SQL SELECT")
             logger.info(self._query(explained_sql))
             answer = input("Continue? [y/n] ")
             if not answer.lower() in ["y", "yes"]:
@@ -107,7 +108,7 @@ class Database(AbstractDatabase):
             assert len(res) == 1, (sql_code, res)
             return res[0]
         elif getattr(res_type, "__origin__", None) is list and len(res_type.__args__) == 1:
-            if res_type.__args__ == (int,) or res_type.__args__ == (str,):
+            if res_type.__args__ in ((int,), (str,)):
                 return [_one(row) for row in res]
             elif res_type.__args__ == (Tuple,):
                 return [tuple(row) for row in res]
@@ -153,9 +154,6 @@ class Database(AbstractDatabase):
         elif issubclass(cls, Decimal):
             if numeric_scale is None:
                 numeric_scale = 0  # Needed for Oracle.
-                # raise ValueError(
-                #     f"{self.name}: Unexpected numeric_scale is NULL, for column {'.'.join(table_path)}.{col_name} of type {type_repr}."
-                # )
             return cls(precision=numeric_scale)
 
         elif issubclass(cls, Float):
@@ -175,20 +173,24 @@ class Database(AbstractDatabase):
         schema, table = self._normalize_table_path(path)
 
         return (
-            "SELECT column_name, data_type, datetime_precision, numeric_precision, numeric_scale FROM information_schema.columns "
+            "SELECT column_name, data_type, datetime_precision, numeric_precision, numeric_scale "
+            "FROM information_schema.columns "
             f"WHERE table_name = '{table}' AND table_schema = '{schema}'"
         )
 
-    def query_table_schema(self, path: DbPath, filter_columns: Optional[Sequence[str]] = None) -> Dict[str, ColType]:
+    def query_table_schema(self, path: DbPath) -> Dict[str, tuple]:
         rows = self.query(self.select_table_schema(path), list)
         if not rows:
             raise RuntimeError(f"{self.name}: Table '{'.'.join(path)}' does not exist, or has no columns")
 
-        if filter_columns is not None:
-            accept = {i.lower() for i in filter_columns}
-            rows = [r for r in rows if r[0].lower() in accept]
+        d = {r[0]: r for r in rows}
+        assert len(d) == len(rows)
+        return d
 
-        col_dict: Dict[str, ColType] = {row[0]: self._parse_type(path, *row) for row in rows}
+    def _process_table_schema(self, path: DbPath, raw_schema: Dict[str, tuple], filter_columns: Sequence[str]):
+        accept = {i.lower() for i in filter_columns}
+
+        col_dict = {row[0]: self._parse_type(path, *row) for name, row in raw_schema.items() if name.lower() in accept}
 
         self._refine_coltypes(path, col_dict)
 
@@ -220,6 +222,25 @@ class Database(AbstractDatabase):
                 else:
                     assert col_name in col_dict
                     col_dict[col_name] = String_UUID()
+                    continue
+
+            alphanum_samples = [s for s in samples if s and String_Alphanum.test_value(s)]
+            if alphanum_samples:
+                if len(alphanum_samples) != len(samples):
+                    logger.warning(
+                        f"Mixed Alphanum/Non-Alphanum values detected in column {'.'.join(table_path)}.{col_name}, disabling Alphanum support."
+                    )
+                else:
+                    assert col_name in col_dict
+                    lens = set(map(len, alphanum_samples))
+                    if len(lens) > 1:
+                        logger.warning(
+                            f"Mixed Alphanum lengths detected in column {'.'.join(table_path)}.{col_name}, disabling Alphanum support."
+                        )
+                    else:
+                        (length,) = lens
+                        col_dict[col_name] = String_Alphanum(length=length)
+                        continue
 
     # @lru_cache()
     # def get_table_schema(self, path: DbPath) -> Dict[str, ColType]:
@@ -248,7 +269,7 @@ class Database(AbstractDatabase):
         return f"concat({joined_exprs})"
 
     def timestamp_value(self, t: DbTime) -> str:
-        return "'%s'" % t.isoformat()
+        return f"'{t.isoformat()}'"
 
     def normalize_uuid(self, value: str, coltype: ColType_UUID) -> str:
         if isinstance(coltype, String_UUID):
