@@ -5,22 +5,16 @@ from datetime import datetime
 
 from runtype import dataclass
 
-from data_diff.utils import ArithUUID
+from data_diff.utils import ArithAlphanumeric, ArithUUID, CaseAwareMapping
 
 
 DbPath = Tuple[str, ...]
-DbKey = Union[int, str, bytes, ArithUUID]
+DbKey = Union[int, str, bytes, ArithUUID, ArithAlphanumeric]
 DbTime = datetime
 
 
 class ColType:
     supported = True
-    pass
-
-
-class IKey(ABC):
-    "Interface for ColType, for using a column as a key in data-diff"
-    python_type: type
 
 
 @dataclass
@@ -63,6 +57,9 @@ class IKey(ABC):
     "Interface for ColType, for using a column as a key in data-diff"
     python_type: type
 
+    def make_value(self, value):
+        return self.python_type(value)
+
 
 class Decimal(FractionalType, IKey):  # Snowflake may use Decimal as a key
     @property
@@ -80,6 +77,10 @@ class ColType_UUID(ColType, IKey):
     python_type = ArithUUID
 
 
+class ColType_Alphanum(ColType, IKey):
+    python_type = ArithAlphanumeric
+
+
 class Native_UUID(ColType_UUID):
     pass
 
@@ -88,6 +89,24 @@ class String_UUID(StringType, ColType_UUID):
     pass
 
 
+@dataclass
+class String_Alphanum(StringType, ColType_Alphanum):
+    length: int
+
+    @staticmethod
+    def test_value(value: str) -> bool:
+        try:
+            ArithAlphanumeric(value)
+            return True
+        except ValueError:
+            return False
+
+    def make_value(self, value):
+        if len(value) != self.length:
+            raise ValueError(f"Expected alphanumeric value of length {self.length}, but got '{value}'.")
+        return self.python_type(value, max_len=self.length)
+
+
 @dataclass
 class Text(StringType):
     supported = False
@@ -121,7 +140,7 @@ class AbstractDatabase(ABC):
         ...
 
     @abstractmethod
-    def concat(self, s: List[str]) -> str:
+    def concat(self, l: List[str]) -> str:
         "Provide SQL for concatenating a bunch of column into a string"
         ...
 
@@ -151,10 +170,23 @@ class AbstractDatabase(ABC):
         ...
 
     @abstractmethod
-    def query_table_schema(self, path: DbPath, filter_columns: Optional[Sequence[str]] = None) -> Dict[str, ColType]:
-        "Query the table for its schema for table in 'path', and return {column: type}"
+    def query_table_schema(self, path: DbPath) -> Dict[str, tuple]:
+        """Query the table for its schema for table in 'path', and return {column: tuple}
+        where the tuple is (table_name, col_name, type_repr, datetime_precision?, numeric_precision?, numeric_scale?)
+        """
         ...
 
+    @abstractmethod
+    def _process_table_schema(self, path: DbPath, raw_schema: Dict[str, tuple], filter_columns: Sequence[str]):
+        """Process the result of query_table_schema().
+
+        Done in a separate step, to minimize the amount of processed columns.
+        Needed because processing each column may:
+        * throw errors and warnings
+        * query the database to sample values
+
+        """
+
     @abstractmethod
     def parse_table_name(self, name: str) -> DbPath:
         "Parse the given table name into a DbPath"
@@ -230,48 +262,9 @@ class AbstractDatabase(ABC):
             return self.normalize_uuid(value, coltype)
         return self.to_string(value)
 
+    @abstractmethod
     def _normalize_table_path(self, path: DbPath) -> DbPath:
         ...
 
 
-class Schema(ABC):
-    @abstractmethod
-    def get_key(self, key: str) -> str:
-        ...
-
-    @abstractmethod
-    def __getitem__(self, key: str) -> ColType:
-        ...
-
-    @abstractmethod
-    def __setitem__(self, key: str, value):
-        ...
-
-    @abstractmethod
-    def __contains__(self, key: str) -> bool:
-        ...
-
-
-class Schema_CaseSensitive(dict, Schema):
-    def get_key(self, key):
-        return key
-
-
-class Schema_CaseInsensitive(Schema):
-    def __init__(self, initial):
-        self._dict = {k.lower(): (k, v) for k, v in dict(initial).items()}
-
-    def get_key(self, key: str) -> str:
-        return self._dict[key.lower()][0]
-
-    def __getitem__(self, key: str) -> ColType:
-        return self._dict[key.lower()][1]
-
-    def __setitem__(self, key: str, value):
-        k = key.lower()
-        if k in self._dict:
-            key = self._dict[k][0]
-        self._dict[k] = key, value
-
-    def __contains__(self, key):
-        return key.lower() in self._dict
+Schema = CaseAwareMapping
