@@ -35,6 +35,10 @@ class Redshift(PostgreSQL):
     def normalize_number(self, value: str, coltype: FractionalType) -> str:
         return self.to_string(f"{value}::decimal(38,{coltype.precision})")
 
+    def concat(self, l: List[str]) -> str:
+        joined_exprs = " || ".join(l)
+        return f"({joined_exprs})"
+
     def select_table_schema(self, path: DbPath) -> str:
         schema, table = self._normalize_table_path(path)
 
