@@ -12,19 +12,28 @@ from concurrent.futures import ThreadPoolExecutor, as_completed
 
 from runtype import dataclass
 
-from .sql import Select, Checksum, Compare, DbPath, DbKey, DbTime, Count, TableName, Time, Value
-from .utils import safezip, split_space
+from .tracking import create_end_event_json, create_start_event_json, send_event_json, is_tracking_enabled
+from .sql import Select, Checksum, Compare, Count, TableName, Time, Value
+from .utils import (
+    CaseAwareMapping,
+    CaseInsensitiveDict,
+    safezip,
+    split_space,
+    CaseSensitiveDict,
+    ArithString,
+    run_as_daemon,
+)
 from .databases.base import Database
 from .databases.database_types import (
-    ArithUUID,
+    DbPath,
+    DbKey,
+    DbTime,
     IKey,
     Native_UUID,
     NumericType,
     PrecisionType,
     StringType,
     Schema,
-    Schema_CaseInsensitive,
-    Schema_CaseSensitive,
 )
 
 logger = logging.getLogger("diff_tables")
@@ -35,12 +44,24 @@ DEFAULT_BISECTION_THRESHOLD = 1024 * 16
 DEFAULT_BISECTION_FACTOR = 32
 
 
+def create_schema(db: Database, table_path: DbPath, schema: dict, case_sensitive: bool) -> CaseAwareMapping:
+    logger.debug(f"[{db.name}] Schema = {schema}")
+
+    if case_sensitive:
+        return CaseSensitiveDict(schema)
+
+    if len({k.lower() for k in schema}) < len(schema):
+        logger.warning(f'Ambiguous schema for {db}:{".".join(table_path)} | Columns = {", ".join(list(schema))}')
+        logger.warning("We recommend to disable case-insensitivity (remove --any-case).")
+    return CaseInsensitiveDict(schema)
+
+
 @dataclass(frozen=False)
 class TableSegment:
     """Signifies a segment of rows (and selected columns) within a table
 
     Parameters:
-        database (Database): Database instance. See :meth:`connect_to_uri`
+        database (Database): Database instance. See :meth:`connect`
         table_path (:data:`DbPath`): Path to table in form of a tuple. e.g. `('my_dataset', 'table_name')`
         key_column (str): Name of the key column, which uniquely identifies each row (usually id)
         update_column (str, optional): Name of updated column, which signals that rows changed (usually updated_at or last_update)
@@ -76,13 +97,15 @@ class TableSegment:
 
     def __post_init__(self):
         if not self.update_column and (self.min_update or self.max_update):
-            raise ValueError("Error: min_update/max_update feature requires to specify 'update_column'")
+            raise ValueError("Error: the min_update/max_update feature requires 'update_column' to be set.")
 
         if self.min_key is not None and self.max_key is not None and self.min_key >= self.max_key:
-            raise ValueError("Error: min_key expected to be smaller than max_key!")
+            raise ValueError(f"Error: min_key expected to be smaller than max_key! ({self.min_key} >= {self.max_key})")
 
         if self.min_update is not None and self.max_update is not None and self.min_update >= self.max_update:
-            raise ValueError("Error: min_update expected to be smaller than max_update!")
+            raise ValueError(
+                f"Error: min_update expected to be smaller than max_update! ({self.min_update} >= {self.max_update})"
+            )
 
     @property
     def _update_column(self):
@@ -116,26 +139,16 @@ class TableSegment:
 
         return self.database.normalize_value_by_type(col, col_type)
 
+    def _with_raw_schema(self, raw_schema: dict) -> "TableSegment":
+        schema = self.database._process_table_schema(self.table_path, raw_schema, self._relevant_columns)
+        return self.new(_schema=create_schema(self.database, self.table_path, schema, self.case_sensitive))
+
     def with_schema(self) -> "TableSegment":
         "Queries the table schema from the database, and returns a new instance of TableSegment, with a schema."
         if self._schema:
             return self
 
-        schema = self.database.query_table_schema(self.table_path, self._relevant_columns)
-        logger.debug(f"[{self.database.name}] Schema = {schema}")
-
-        schema_inst: Schema
-        if self.case_sensitive:
-            schema_inst = Schema_CaseSensitive(schema)
-        else:
-            if len({k.lower() for k in schema}) < len(schema):
-                logger.warning(
-                    f'Ambiguous schema for {self.database}:{".".join(self.table_path)} | Columns = {", ".join(list(schema))}'
-                )
-                logger.warning("We recommend to disable case-insensitivity (remove --any-case).")
-            schema_inst = Schema_CaseInsensitive(schema)
-
-        return self.new(_schema=schema_inst)
+        return self._with_raw_schema(self.database.query_table_schema(self.table_path))
 
     def _make_key_range(self):
         if self.min_key is not None:
@@ -175,10 +188,10 @@ class TableSegment:
     def choose_checkpoints(self, count: int) -> List[DbKey]:
         "Suggests a bunch of evenly-spaced checkpoints to split by (not including start, end)"
         assert self.is_bounded
-        if isinstance(self.min_key, ArithUUID):
+        if isinstance(self.min_key, ArithString):
+            assert type(self.min_key) is type(self.max_key)
             checkpoints = split_space(self.min_key.int, self.max_key.int, count)
-            assert isinstance(self.max_key, ArithUUID)
-            return [ArithUUID(int=i) for i in checkpoints]
+            return [self.min_key.new(int=i) for i in checkpoints]
 
         return split_space(self.min_key, self.max_key, count)
 
@@ -221,11 +234,11 @@ class TableSegment:
 
     def count_and_checksum(self) -> Tuple[int, int]:
         """Count and checksum the rows in the segment, in one pass."""
-        start = time.time()
+        start = time.monotonic()
         count, checksum = self.database.query(
             self._make_select(columns=[Count(), Checksum(self._relevant_columns_repr)]), tuple
         )
-        duration = time.time() - start
+        duration = time.monotonic() - start
         if duration > RECOMMENDED_CHECKSUM_DURATION:
             logger.warning(
                 f"Checksum is taking longer than expected ({duration:.2f}s). "
@@ -256,6 +269,11 @@ class TableSegment:
     def is_bounded(self):
         return self.min_key is not None and self.max_key is not None
 
+    def approximate_size(self):
+        if not self.is_bounded:
+            raise RuntimeError("Cannot approximate the size of an unbounded segment. Must have min_key and max_key.")
+        return self.max_key - self.min_key
+
 
 def diff_sets(a: set, b: set) -> Iterator:
     s1 = set(a)
@@ -268,7 +286,7 @@ def diff_sets(a: set, b: set) -> Iterator:
     for i in s2 - s1:
         d[i[0]].append(("+", i))
 
-    for k, v in sorted(d.items(), key=lambda i: i[0]):
+    for _k, v in sorted(d.items(), key=lambda i: i[0]):
         yield from v
 
 
@@ -311,8 +329,8 @@ class TableDiffer:
 
         Returns:
             An iterator that yield pair-tuples, representing the diff. Items can be either
-            ('+', columns) for items in table1 but not in table2
-            ('-', columns) for items in table2 but not in table1
+            ('-', columns) for items in table1 but not in table2
+            ('+', columns) for items in table2 but not in table1
             Where `columns` is a tuple of values for the involved columns, i.e. (id, ...extra)
         """
         # Validate options
@@ -321,49 +339,83 @@ class TableDiffer:
         if self.bisection_factor < 2:
             raise ValueError("Must have at least two segments per iteration (i.e. bisection_factor >= 2)")
 
-        # Query and validate schema
-        table1, table2 = self._threaded_call("with_schema", [table1, table2])
-        self._validate_and_adjust_columns(table1, table2)
+        if is_tracking_enabled():
+            options = dict(self)
+            event_json = create_start_event_json(options)
+            run_as_daemon(send_event_json, event_json)
 
-        key_type = table1._schema[table1.key_column]
-        key_type2 = table2._schema[table2.key_column]
-        if not isinstance(key_type, IKey):
-            raise NotImplementedError(f"Cannot use column of type {key_type} as a key")
-        if not isinstance(key_type2, IKey):
-            raise NotImplementedError(f"Cannot use column of type {key_type2} as a key")
-        assert key_type.python_type is key_type2.python_type
+        self.stats["diff_count"] = 0
+        start = time.monotonic()
+        try:
 
-        # Query min/max values
-        key_ranges = self._threaded_call_as_completed("query_key_range", [table1, table2])
+            # Query and validate schema
+            table1, table2 = self._threaded_call("with_schema", [table1, table2])
+            self._validate_and_adjust_columns(table1, table2)
 
-        # Start with the first completed value, so we don't waste time waiting
-        min_key1, max_key1 = self._parse_key_range_result(key_type, next(key_ranges))
+            key_type = table1._schema[table1.key_column]
+            key_type2 = table2._schema[table2.key_column]
+            if not isinstance(key_type, IKey):
+                raise NotImplementedError(f"Cannot use column of type {key_type} as a key")
+            if not isinstance(key_type2, IKey):
+                raise NotImplementedError(f"Cannot use column of type {key_type2} as a key")
+            assert key_type.python_type is key_type2.python_type
 
-        table1, table2 = [t.new(min_key=min_key1, max_key=max_key1) for t in (table1, table2)]
+            # Query min/max values
+            key_ranges = self._threaded_call_as_completed("query_key_range", [table1, table2])
 
-        logger.info(
-            f"Diffing tables | segments: {self.bisection_factor}, bisection threshold: {self.bisection_threshold}. "
-            f"key-range: {table1.min_key}..{table2.max_key}, "
-            f"size: {table2.max_key-table1.min_key}"
-        )
+            # Start with the first completed value, so we don't waste time waiting
+            min_key1, max_key1 = self._parse_key_range_result(key_type, next(key_ranges))
 
-        # Bisect (split) the table into segments, and diff them recursively.
-        yield from self._bisect_and_diff_tables(table1, table2)
+            table1, table2 = [t.new(min_key=min_key1, max_key=max_key1) for t in (table1, table2)]
 
-        # Now we check for the second min-max, to diff the portions we "missed".
-        min_key2, max_key2 = self._parse_key_range_result(key_type, next(key_ranges))
+            logger.info(
+                f"Diffing tables | segments: {self.bisection_factor}, bisection threshold: {self.bisection_threshold}. "
+                f"key-range: {table1.min_key}..{table2.max_key}, "
+                f"size: {table1.approximate_size()}"
+            )
 
-        if min_key2 < min_key1:
-            pre_tables = [t.new(min_key=min_key2, max_key=min_key1) for t in (table1, table2)]
-            yield from self._bisect_and_diff_tables(*pre_tables)
+            # Bisect (split) the table into segments, and diff them recursively.
+            yield from self._bisect_and_diff_tables(table1, table2)
 
-        if max_key2 > max_key1:
-            post_tables = [t.new(min_key=max_key1, max_key=max_key2) for t in (table1, table2)]
-            yield from self._bisect_and_diff_tables(*post_tables)
+            # Now we check for the second min-max, to diff the portions we "missed".
+            min_key2, max_key2 = self._parse_key_range_result(key_type, next(key_ranges))
+
+            if min_key2 < min_key1:
+                pre_tables = [t.new(min_key=min_key2, max_key=min_key1) for t in (table1, table2)]
+                yield from self._bisect_and_diff_tables(*pre_tables)
+
+            if max_key2 > max_key1:
+                post_tables = [t.new(min_key=max_key1, max_key=max_key2) for t in (table1, table2)]
+                yield from self._bisect_and_diff_tables(*post_tables)
+
+            error = None
+        except BaseException as e:  # Catch KeyboardInterrupt too
+            error = e
+        finally:
+            if is_tracking_enabled():
+                runtime = time.monotonic() - start
+                table1_count = self.stats.get("table1_count")
+                table2_count = self.stats.get("table2_count")
+                diff_count = self.stats.get("diff_count")
+                err_message = str(error)[:20]  # Truncate possibly sensitive information.
+                event_json = create_end_event_json(
+                    error is None,
+                    runtime,
+                    table1.database.name,
+                    table2.database.name,
+                    table1_count,
+                    table2_count,
+                    diff_count,
+                    err_message,
+                )
+                send_event_json(event_json)
+
+            if error:
+                raise error
 
     def _parse_key_range_result(self, key_type, key_range):
         mn, mx = key_range
-        cls = key_type.python_type
+        cls = key_type.make_value
         # We add 1 because our ranges are exclusive of the end (like in Python)
         try:
             return cls(mn), cls(mx) + 1
@@ -437,6 +489,8 @@ class TableDiffer:
                 self.stats["table1_count"] = len(rows1)
                 self.stats["table2_count"] = len(rows2)
 
+            self.stats["diff_count"] += len(diff)
+
             logger.info(". " * level + f"Diff found {len(diff)} different rows.")
             self.stats["rows_downloaded"] = self.stats.get("rows_downloaded", 0) + max(len(rows1), len(rows2))
             yield from diff
